#include "AntTweakBarPlugin.h"
#include <malloc.h>
#include <math.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <GL/GL.h>
//#include <GL/GLU.h>

//#define USEDEBUG

#ifdef USEDEBUG
#define MSG(msg) MessageBox(NULL,__FILE__,msg,MB_ICONERROR);
#else
#define MSG(msg)
#endif

#include <Basic4GLDLLInterface.h>
#include <Basic4GLStandardObjects.h>
#include "Basic4GLPluginMacros.h"
#include "AntTweakBar.h"

const size_t MAX_CHARS = 512;
const float EPSILON = 0.000001f;

struct BarVar{
  TwBar * Bar;
  char *  VarName;
  ETwType VarType;
  union {
    int ButtonCode;
    int SetCode;
  };
  int     GetCode;
  int     UserData;
  int     Index;
  union {
    void * pVar;
    char * pCDString;
    int  * pInt;
    float * pFloat;
  };
};

// global variables.
char * pluginError = NULL;
IB4GLCompiler * gpCompiler = NULL;
int IsInit = 0;

int nVars = 0;
BarVar ** pVars=NULL;

char * strDup(const char * s) {
  size_t len = strlen(s);
  char * p = (char*)malloc(len+1);
  memcpy(p, s, len);
  p[len]=0;
  return p;
}

int AllocVar(void) {
  int index = -1;
  if (nVars>0) {
    for (int i=0; i<nVars; i++){
      if (!pVars[i]->Bar){
        index=i;
        break;
      }
    }
  }
  if (index==-1) {
    index=nVars;
    nVars++;
    pVars = (BarVar**)realloc(pVars,sizeof(BarVar*) * nVars);
    pVars[index] = (BarVar*)malloc(sizeof(BarVar));
    memset(pVars[index], 0, sizeof(BarVar));
  }
  if (pVars[index]->VarName) {
    free(pVars[index]->VarName);
    pVars[index]->VarName=NULL;
  }
  if (pVars[index]->pVar) {
    free(pVars[index]->pVar);
    pVars[index]->pVar=NULL;
  }
  pVars[index]->Index=index;
  pVars[index]->Bar=NULL;
  pVars[index]->SetCode=0;
  pVars[index]->GetCode=0;
  pVars[index]->UserData=0;

  return index;
}

void FreeAllVars(void) {
  if (!pVars) return;
  if (nVars>0) {
    for(int index=0;index<nVars;index++){
      if (pVars[index]){
        if (pVars[index]->VarName) {
          free(pVars[index]->VarName);
          pVars[index]->VarName=NULL;
        }
        if (pVars[index]->pVar) {
          free(pVars[index]->pVar);
          pVars[index]->pVar=NULL;
        }
        free(pVars[index]);
        pVars[index]=NULL;
      }
    }
  }
  nVars=0;
  free(pVars);
  pVars=NULL;
}

void TW_CALL CopyCDStringToClient(char **destPtr, const char *src){
  //MSG("CopyCDStringToClient");
  size_t srcLen = (src!=NULL) ? strlen(src) : 0;
  size_t destLen = (*destPtr!=NULL) ? strlen(*destPtr) : 0;
 
  // alloc or realloc dest memory block if needed
  if( *destPtr==NULL )
    *destPtr = (char *)malloc(srcLen+1);
  else if( srcLen>destLen )
    *destPtr = (char *)realloc(*destPtr, srcLen+1);
 
  // copy src
  if( srcLen>0 )
    strncpy(*destPtr, src, srcLen);
  (*destPtr)[srcLen] = '\0'; // null-terminated string
}

void TW_CALL ButtonCallback(void *clientData) { 
  BarVar * barvar = (BarVar *)clientData;
  if (barvar) {
    if (barvar->ButtonCode)
      gpCompiler->Execute(barvar->ButtonCode);
  }
}

void TW_CALL SetVarCallback(const void *value, void *clientData){
  BarVar * barvar = (BarVar *)clientData;
  if (!barvar) return;

  if (barvar->SetCode) {
    void * v = (void*) value;
    switch (barvar->VarType){
    case TW_TYPE_BOOL32 :memcpy(barvar->pInt,  v, 4); break;
    case TW_TYPE_INT32  :memcpy(barvar->pInt,  v, 4); break;
    case TW_TYPE_COLOR32:memcpy(barvar->pInt,  v, 4); break;
    case TW_TYPE_FLOAT  :memcpy(barvar->pFloat,v, 4); break;
    case TW_TYPE_DIR3F  :memcpy(barvar->pFloat,v,12); break;
    case TW_TYPE_COLOR3F:memcpy(barvar->pFloat,v,12); break;
    case TW_TYPE_COLOR4F:memcpy(barvar->pFloat,v,16); break;
    case TW_TYPE_QUAT4F :memcpy(barvar->pFloat,v,16); break;
    case TW_TYPE_CDSTRING: 
      const char *src = *(const char **)(value);
      if (barvar->pCDString)
        free(barvar->pCDString);
      barvar->pCDString = strDup(src);
      break;
    }
    gpCompiler->Execute(barvar->SetCode);
  }
}

void TW_CALL GetVarCallback(void *v, void *clientData){
  BarVar * barvar = (BarVar *)clientData;
  if (!barvar) return;
  if (barvar->GetCode) {
    gpCompiler->Execute(barvar->GetCode);
    switch (barvar->VarType){
    case TW_TYPE_BOOL32  : memcpy(v,barvar->pInt  , 4); break;
    case TW_TYPE_INT32   : memcpy(v,barvar->pInt  , 4); break;
    case TW_TYPE_COLOR32 : memcpy(v,barvar->pInt  , 4); break;
    case TW_TYPE_FLOAT   : memcpy(v,barvar->pFloat, 4); break;
    case TW_TYPE_DIR3F   : memcpy(v,barvar->pFloat,12); break;
    case TW_TYPE_COLOR3F : memcpy(v,barvar->pFloat,12); break;
    case TW_TYPE_COLOR4F : memcpy(v,barvar->pFloat,16); break;
    case TW_TYPE_QUAT4F  : memcpy(v,barvar->pFloat,16); break;
    case TW_TYPE_CDSTRING: 
      char **destPtr = (char **)v;
      TwCopyCDStringToLibrary(destPtr, barvar->pCDString);
      break;
    }
  }
}

int AllocBarVarCB(TwBar* Bar,const char * VarName,ETwType VarType,const char * strSetCode, const char * strGetCode, const char * Define){
  if (!IsInit) return -1;
  if (TwGetBarCount()<1) return -1;
  if (!Bar) return -1;
  if (!VarName) return -1;
  if (!strlen(VarName)) return -1;
  if (!strGetCode) return -1;
  if (!strlen(strGetCode)) return -1;

  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_COLOR32:
  case TW_TYPE_FLOAT:
  case TW_TYPE_CDSTRING:
  case TW_TYPE_DIR3F:
  case TW_TYPE_COLOR3F:
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:
    break;
  default:
    return -1;
  }
  
  char code[256];

  int index=123; // fake

  int SetCode=0;
  if (strSetCode) {
    sprintf(code,"%s( %d )",strSetCode,index);
    SetCode=gpCompiler->Compile(code);
    if (!SetCode) return -1;
  }
  
  sprintf(code,"%s( %d )",strGetCode,index);
  int GetCode=gpCompiler->Compile(code);
  if (!GetCode) return -1;


  index=AllocVar(); // the real index
  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_COLOR32:
  case TW_TYPE_FLOAT:    pVars[index]->pVar=malloc(4); break;
  case TW_TYPE_CDSTRING: pVars[index]->pVar=malloc(1); pVars[index]->pCDString[0]=0; break;
  case TW_TYPE_DIR3F:
  case TW_TYPE_COLOR3F:  pVars[index]->pVar=malloc(12); break;
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:   pVars[index]->pVar=malloc(16); break;
  }
  if (strSetCode) {
    sprintf(code,"%s( %d )",strSetCode,index);
    SetCode=gpCompiler->Compile(code);
  }
  sprintf(code,"%s( %d )",strGetCode,index);
  GetCode=gpCompiler->Compile(code);

  pVars[index]->Bar      = Bar;
  pVars[index]->VarName  = strDup(VarName);
  pVars[index]->VarType  = VarType;
  pVars[index]->SetCode  = SetCode;
  pVars[index]->GetCode  = GetCode;
  
  if (!SetCode)
    TwAddVarCB(Bar,VarName,VarType,NULL           ,&GetVarCallback,pVars[index],Define);
  else
    TwAddVarCB(Bar,VarName,VarType,&SetVarCallback,&GetVarCallback,pVars[index],Define);

  return index;
}


int AllocBarVar(bool ReadOnly, TwBar* Bar, const char * VarName, ETwType VarType, const void * v,const char * Define){
  if (!IsInit) return -1;
  if (TwGetBarCount()<1) return -1;
  if (!Bar) return -1;
  if (!VarName) return -1;
  if (!strlen(VarName)) return -1;
  if (!v) return -1;

  int nBytes;
  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_FLOAT:
  case TW_TYPE_COLOR32: nBytes=4; break;
  case TW_TYPE_DIR3F:
  case TW_TYPE_COLOR3F: nBytes=12; break; 
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:  nBytes=16; break; 
  case TW_TYPE_CDSTRING: nBytes=strlen((char*)v)+1; break;
  default:
    return -1;
  }

  int index=AllocVar();

  pVars[index]->Bar = Bar;
  pVars[index]->VarName = strDup(VarName);
  pVars[index]->VarType = VarType;
  pVars[index]->pVar = malloc(nBytes);
  memcpy(pVars[index]->pVar,v,nBytes);
  
  void * linked;
  if (VarType==TW_TYPE_CDSTRING)
    linked = &pVars[index]->pVar;
  else
    linked = pVars[index]->pVar;

  if (ReadOnly)
    TwAddVarRO(pVars[index]->Bar, pVars[index]->VarName, VarType, linked, Define);
  else
    TwAddVarRW(pVars[index]->Bar, pVars[index]->VarName, VarType, linked, Define);
    
  return index;
}


void FreeBarVar(TwBar* Bar, const char * VarName){
  if (!pVars) return;
  if (nVars<1) return;
  if (!Bar) return;
  if (!VarName) return;
  if (!strlen(VarName)) return;

  for(int index=0;index<nVars;index++){
    if (pVars[index]->Bar==Bar) {
      if (pVars[index]->VarName) {
        if (strcmp(pVars[index]->VarName,VarName)==0) {
          free(pVars[index]->VarName);
          pVars[index]->VarName=NULL;
          if (pVars[index]->pVar) {
            free(pVars[index]->pVar);
            pVars[index]->pVar=NULL;
          }
          pVars[index]->Bar=NULL;
          break;
        }
      }
    }
  }
}

void FreeAllBarVars(TwBar* Bar){
  if (!pVars) return;
  if (nVars<1) return;
  if (!Bar) return;
  for(int index=0;index<nVars;index++){
    if (pVars[index]->Bar==Bar) {
      if (pVars[index]->VarName) {
        free(pVars[index]->VarName);
        pVars[index]->VarName=NULL;
      }
      if (pVars[index]->pVar) {
        free(pVars[index]->pVar);
        pVars[index]->pVar=NULL;
      }
      pVars[index]->Bar=NULL;
    }
  }
}


//HWND hWin = NULL;
TwBar * pBar = NULL;
int hBar = 0;

int WindowWidth = 0;
int WindowHeight = 0;


class AntTweakBarPlugin : public IDLL_Basic4GL_Plugin {
  public:
  AntTweakBarPlugin() { ; }
  virtual bool DLLFUNC Load(IDLL_Basic4GL_FunctionRegistry &registry, bool isStandaloneExe);
  virtual bool DLLFUNC Start();
  virtual void DLLFUNC End();
  virtual void DLLFUNC Unload();
  virtual void DLLFUNC GetError(char *error);
  virtual void DLLFUNC Pause();
  virtual void DLLFUNC Resume();
  virtual void DLLFUNC DelayedResume();
  virtual void DLLFUNC ProcessMessages();
};


#define GBAR(n) GT(hBar,n,&pBar)
#define SBAR(n) ST(hBar,n,&pBar)
#define RBAR() STR(hBar,&pBar)

// #######
// # API #
// #######
DPF(TwGetLastError){
  const char * p = TwGetLastError();
  if (!p)
    p = &"";
  SSR(p);
}
DPF(TwInit){ // result%=TwInit(WindowWidth%,WindowHeight%)
  if (!IsInit){
    TwCopyCDStringToClientFunc(&CopyCDStringToClient);
    IsInit=TwInit(TW_OPENGL,NULL);
    if (IsInit) {
      WindowWidth=GI(2); WindowHeight=GI(1);
      TwWindowSize(WindowWidth,WindowHeight);
    }
  }
  SIR(IsInit);
}

DPF(TwIsInit){ // result%=TwIsInit()
  SIR(IsInit);
}

DPF(TwTerminate){ // result% = TwTerminate()
  SIR(TwTerminate());
  IsInit=0;
}

DPF(TwWindowSize){ // result%=TwWindowSize(Width%,Height%)
  WindowWidth=GI(2); WindowHeight=GI(1);
  SIR(TwWindowSize(WindowWidth,WindowHeight));
}
DPF(TwMouseMotion){ // result%=TwMouseMotion(MouseX#,MouseY#)
  if (!IsInit || TwGetBarCount()<1) { SIR(0); return; }
  float x = GF(2) * (float)WindowWidth;
  float y = GF(1) * (float)WindowHeight;
  SIR(TwMouseMotion((int)x,(int)y));
}

DPF(TwMouseButton){ // result%=TwMouseButton(TW_ACTION%,TW_BUTTON%)
  if (!IsInit || TwGetBarCount()<1) { SIR(0); return; }
  SIR(TwMouseButton((ETwMouseAction)GI(2),( ETwMouseButtonID)GI(1)));
}

DPF(TwMouseButtons){ // result%=TwMouseButtons(LeftButton%,MiddleButton%,Rightbutton%)
  static int oldLButton=0;
  static int oldMButton=0;
  static int oldRButton=0;
  SIR(0); if (!IsInit || TwGetBarCount()<1) return; 

  int Button=GI(3);
  if (Button!=oldLButton) {
    if (oldLButton==0)
      TwMouseButton(TW_MOUSE_PRESSED,TW_MOUSE_LEFT);
    else
      TwMouseButton(TW_MOUSE_RELEASED,TW_MOUSE_LEFT);
    oldLButton=Button;
  }
  Button=GI(2);
  if (Button!=oldMButton) {
    if (oldMButton==0)
      TwMouseButton(TW_MOUSE_PRESSED,TW_MOUSE_MIDDLE);
    else
      TwMouseButton(TW_MOUSE_RELEASED,TW_MOUSE_MIDDLE);
    oldMButton=Button;

  }
  Button=GI(1);
  if (Button!=oldRButton) {
    if (oldRButton==0)
      TwMouseButton(TW_MOUSE_PRESSED,TW_MOUSE_RIGHT);
    else
      TwMouseButton(TW_MOUSE_RELEASED,TW_MOUSE_RIGHT);
    oldRButton=Button;
  }
  SIR(1);
}
DPF(TwMouseWheel){
  static int oldWheel=0;
  if (!IsInit || TwGetBarCount()<1) { SIR(0); return; }
  int wheel=oldWheel+GI(1);
  SIR(TwMouseWheel(wheel));
  oldWheel=wheel;
}

DPF(TwKeyPressed){ // result%=TwKeyPressed(TW_KEY%,TW_MODIFIER%)
  if (!IsInit || TwGetBarCount()<1) { SIR(0); return; }
  SIR(TwKeyPressed(GI(2),GI(1)));
}


DPF(TwDefine){
  SIR(TwDefine(GS(1)));
}

DPF(TwScanKey){ // result%=TwScanKey(ScanKey%)
  static int oldModifier = TW_KMOD_NONE;
  SIR(0);
  if (!IsInit || TwGetBarCount()<1) return;

  int scan = GI(1);
  if (!scan) return;  
  
  int key=0;
  int modifier = oldModifier;

  if (scan==8)
    key = TW_KEY_BACKSPACE;
  else if (scan==9)
    key = TW_KEY_TAB;
  else if (scan==13)
    key = TW_KEY_RETURN;
    
  else if (scan==16)
    modifier |= TW_KMOD_SHIFT;
  else if (scan==17)
    modifier |= TW_KMOD_CTRL;
  else if (scan==18)
    modifier |= TW_KMOD_ALT;

  else if (scan==27)
    key = TW_KEY_ESCAPE;

  else if (scan==32)
    key = TW_KEY_SPACE;
  else if (scan==33)
    key = TW_KEY_PAGE_UP;
  else if (scan==34)
    key = TW_KEY_PAGE_DOWN;
  else if (scan==35)
    key = TW_KEY_END; 
  else if (scan==36)
    key = TW_KEY_HOME;
  else if (scan==37)
    key = TW_KEY_LEFT;
  else if (scan==38)
    key = TW_KEY_UP;
  else if (scan==39)
    key = TW_KEY_RIGHT;
  else if (scan==40)
    key = TW_KEY_DOWN;

  else if (scan==45)
    key = TW_KEY_INSERT;
  else if (scan==46)
    key = TW_KEY_DELETE;   

  else if (scan>47 && scan<58)
    key = scan; // 0-9

  else if (scan>64 && scan<91)
  {
    key = scan; // A-Z
    key+=32;
    if (modifier && TW_KMOD_SHIFT) {
      key-=32;
      modifier = 0;
    }
    
  }
  else if (scan>90 && scan<94)
    modifier |= TW_KMOD_META;
    
  else if (scan>95 && scan<112)
    key = scan; // number block

  else if (scan>111 && scan<124)
    key = TW_KEY_F1 + (scan-112); // F1 - F12
  else if (scan==-66)
    key = '.';
  else if (scan==-68)
    key = ',';
  
  if (!key)
    oldModifier = modifier;
  else{
    SIR(TwKeyPressed(key,modifier));
    oldModifier=TW_KMOD_NONE;
  }
}

DPF(TwDraw){ // resul%=TwDraw()
  if (!IsInit || TwGetBarCount()<1) {
    SIR(0); return;
  }
  SIR(TwDraw());
}

// ############
// # Tweakbar #
// ############
DPF(TwNewBar){ // bar = TwNewBar(name$)
  pBar=TwNewBar(GS(1)); RBAR();
}
DPF(TwDeleteBar){ // result% = TwDeleteBar(bar)
  GBAR(1);
  if (pBar) { 
    SIR(TwDeleteBar(pBar));  
    FreeAllBarVars(pBar);
    pBar=0; SBAR(1); 
  } 
  else
    SIR(0);
}

DPF(TwDeleteAllBars){ // resul%=TwDeleteAllBars()
  SIR(TwDeleteAllBars);
  FreeAllVars();
}

DPF(TwSetTopBar){ // result% = TwSetTopBar(Bar)
  GBAR(1); if (pBar) SIR(TwSetTopBar(pBar)); else SIR(0);
}
DPF(TwGetTopBar){ // bar = TwGetTopBar()
  pBar=TwGetTopBar();RBAR();
}
  
DPF(TwSetBottomBar){ // result% = TwSetBottomBar(Bar)
  GBAR(1); if (pBar) SIR(TwSetBottomBar(pBar)); else SIR(0);
}
DPF(TwGetBottomBar){ // bar = TwGetBottomBar()
  pBar=TwGetBottomBar();RBAR();
}  

DPF(TwGetBarName){ // name$ = TwGetBarName(Bar)
  GBAR(1); if (pBar) SSR(TwGetBarName(pBar)); else SSR("");
}
DPF(TwGetBarByName){ // bar = TwGetBarByName(name$)
  pBar=TwGetBarByName(GS(1));RBAR();
}

DPF(TwGetBarCount){ // count% = TwGetBarCount()
  SIR(TwGetBarCount());
}
DPF(TwGetBarByIndex){ // bar = TwGetBarByIndex(index%)
  pBar=TwGetBarByIndex(GI(1));RBAR();
}
  
DPF(TwRefreshBar){ // result% = TwRefreshBar(Bar)
  GBAR(1); if (pBar) SIR(TwRefreshBar(pBar)); else SIR(0);
}

// ##########
// # Butons #
// ##########
DPF(TwAddButtonD){ // index%=TwAddButton(bar,ButtonName$,ButtonCallback$,Define$)
  GBAR(4); if(!pBar){SIR(-1); return;}
  int index=123; // fake
  char code[256];
  sprintf(code,"%s( %d )",GS(2),index);
  int ButtonCode=gpCompiler->Compile(code);
  if (!ButtonCode) {SIR(-1); return;}

  index = AllocVar(); // real index
  sprintf(code,"%s( %d )",GS(2),index);
  ButtonCode=gpCompiler->Compile(code);

  pVars[index]->Bar=pBar;
  pVars[index]->VarName=strDup(GS(3));
  pVars[index]->VarType=TW_TYPE_UNDEF;
  pVars[index]->ButtonCode=ButtonCode;
  
  TwAddButton(pVars[index]->Bar,pVars[index]->VarName,&ButtonCallback,pVars[index],GS(1));
  SIR(index);
}
DPF(TwAddButton){ // index%=TwAddButton(bar,ButtonName$,ButtonCallback$)
  GBAR(3); if(!pBar){SIR(-1); return;}
  int index=123; // fake
  char code[256];
  sprintf(code,"%s( %d )",GS(1),index);
  int ButtonCode=gpCompiler->Compile(code);
  if (!ButtonCode) {SIR(-1); return;}

  index = AllocVar(); // real index
  sprintf(code,"%s( %d )",GS(1),index);
  ButtonCode=gpCompiler->Compile(code);

  pVars[index]->Bar=pBar;
  pVars[index]->VarName=strDup(GS(2));
  pVars[index]->VarType=TW_TYPE_UNDEF;
  pVars[index]->ButtonCode=ButtonCode;
  
  TwAddButton(pVars[index]->Bar,pVars[index]->VarName,&ButtonCallback,pVars[index],NULL);
  SIR(index);
}

  // #############
  // # Separator #
  // #############
DPF(TwAddSeparatorD){ // Index%=TwAddSeparator(bar, SeparatorName$, Define$)
  GBAR(3); if(!pBar){SIR(-1); return;}
  int index = AllocVar();
  pVars[index]->Bar=pBar;
  pVars[index]->VarName=strDup(GS(2));
  pVars[index]->VarType=TW_TYPE_UNDEF;
  TwAddSeparator(pVars[index]->Bar,pVars[index]->VarName,GS(1));
}

DPF(TwAddSeparator){ // Index%=TwAddSeparator(bar, SeparatorName$)
  GBAR(2); if(!pBar){SIR(-1); return;}
  int index = AllocVar();
  pVars[index]->Bar=pBar;
  pVars[index]->VarName=strDup(GS(1));
  pVars[index]->VarType=TW_TYPE_UNDEF;
  TwAddSeparator(pVars[index]->Bar,pVars[index]->VarName,NULL);
}

// #################
// # Callback Vars #
// #################
DPF(TwAddVarCBD){   // index% = TwAddVarCB(bar,VarName$,VarType%,SetCallback$,GetCallback$,Define$)
  GBAR(6); SIR(AllocBarVarCB(pBar,GS(5),(ETwType)GI(4), GS(3),GS(2), GS(1)));
}
DPF(TwAddVarCB){   // index% = TwAddVarCB(bar,VarName$,VarType%,SetCallback$,GetCallback$)
  GBAR(5); SIR(AllocBarVarCB(pBar,GS(4),(ETwType)GI(3), GS(2),GS(1), NULL));
}
DPF(TwAddVarROCBD){ // index% = TwAddVarROCB(bar,VarName$,VarType%,GetCallback$,Define$)
  GBAR(5); SIR(AllocBarVarCB(pBar,GS(4),(ETwType)GI(3), NULL,GS(2), GS(1)));
}
DPF(TwAddVarROCB){ // index% = TwAddVarROCB(bar,VarName$,VarType%,GetCallback$)
  GBAR(4); SIR(AllocBarVarCB(pBar,GS(3),(ETwType)GI(2), NULL,GS(1), NULL));
}
// ###########
// # RO Vars #
// ###########
DPF(TwAddVarROD) { // index% = TwAddVarRO(bar,VarName$,VarType%,Value,Define$)
  int   Int;
  float Floats[16];
  GBAR(5); ETwType VarType = (ETwType)GI(3);
  void * v;
  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_COLOR32:  v=&Int; Int=GI(2); break;
  case TW_TYPE_FLOAT:    v=&Floats; Floats[0]=GF(2); break;
  case TW_TYPE_CDSTRING: v=(void*)GS(2); break;
  case TW_TYPE_DIR3F:    
  case TW_TYPE_COLOR3F:  v=&Floats; GFA(2,(float*)v,3); break;
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:   v=&Floats; GFA(2,(float*)v,4); break; 
  }
  SIR(AllocBarVar(true,pBar,GS(4),VarType,v,GS(1)));
}

DPF(TwAddVarRO) { // index% = TwAddVarRO(bar,VarName$,VarType%,Value)
  int   Int;
  float Floats[16];
  GBAR(4); ETwType VarType = (ETwType)GI(2);
  void * v;
  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_COLOR32:  v=&Int; Int=GI(1); break;
  case TW_TYPE_FLOAT:    v=&Floats; Floats[0]=GF(1); break;
  case TW_TYPE_CDSTRING: v=(void*)GS(1); break;
  case TW_TYPE_DIR3F:    
  case TW_TYPE_COLOR3F:  v=&Floats; GFA(1,(float*)v,3); break;
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:   v=&Floats; GFA(1,(float*)v,4); break; 
  }
  SIR(AllocBarVar(true,pBar,GS(3),VarType,v,NULL));
}

// ###########
// # RW Vars #
// ###########
DPF(TwAddVarRWD) { // index% = TwAddVarRW(bar,VarName$,VarType%,Value,Define$)
  int   Int;
  float Floats[16];
  GBAR(5); ETwType VarType = (ETwType)GI(3);
  void * v;
  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_COLOR32:  v=&Int; Int=GI(2); break;
  case TW_TYPE_FLOAT:    v=&Floats; Floats[0]=GF(2); break;
  case TW_TYPE_CDSTRING: v=(void*)GS(2); break;
  case TW_TYPE_DIR3F:    
  case TW_TYPE_COLOR3F:  v=&Floats; GFA(2,(float*)v,3); break;
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:   v=&Floats; GFA(2,(float*)v,4); break; 
  }
  SIR(AllocBarVar(false,pBar,GS(4),(ETwType)GI(3),v,GS(1)));
}

DPF(TwAddVarRW) { // index% = TwAddVarRW(bar,VarName$,VarType%,Value)
  int   Int;
  float Floats[16];
  GBAR(4); ETwType VarType = (ETwType)GI(2);
  void * v;
  switch (VarType){
  case TW_TYPE_BOOL32:
  case TW_TYPE_INT32:
  case TW_TYPE_COLOR32:  v=&Int; Int=GI(1); break;
  case TW_TYPE_FLOAT:    v=&Floats; Floats[0]=GF(1); break;
  case TW_TYPE_CDSTRING: v=(void*)GS(1); break;
  case TW_TYPE_DIR3F:    
  case TW_TYPE_COLOR3F:  v=&Floats; GFA(1,(float*)v,3); break;
  case TW_TYPE_COLOR4F:
  case TW_TYPE_QUAT4F:   v=&Floats; GFA(1,(float*)v,4); break; 
  }
  SIR(AllocBarVar(false,pBar,GS(3),VarType,v,NULL));
}

// ########
// # read #
// ########
DPF(TwReadBool32){ // integer%=TwReadBool32(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SIR(*pVars[index]->pInt);
}

DPF(TwReadInt32){ // integer%=TwReadInt32(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SIR(*pVars[index]->pInt);
}

DPF(TwReadColor32){ // integer%=TwReadColor32(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SIR(*pVars[index]->pInt);
}

DPF(TwReadFloat){ // float#=TwReadFloat(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SFR(*pVars[index]->pFloat);
}

DPF(TwReadDir3f){ // dir#()=TwReadDir3f(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SFAR(pVars[index]->pFloat,3);
}

DPF(TwReadColor3f){ // color#()=TwReadColor3f(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SFAR(pVars[index]->pFloat,3);
}

DPF(TwReadColor4f){ // color#()=TwReadColor4f(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SFAR(pVars[index]->pFloat,4);
}

DPF(TwReadQuat4f){ // quat#()=TwReadQuat4f(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SFAR(pVars[index]->pFloat,4);
}

DPF(TwReadCDString){ // string$=TwReadCDString(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SSR((const char*)pVars[index]->pCDString);
}

// #########
// # write #
// #########
DPF(TwWriteBool32){ // TwWriteBool32(index%,value%)
  int index=GI(2); if (index<0 || index>=nVars) return;
  *pVars[index]->pInt=GI(1);
}
DPF(TwWriteInt32){ // TwWriteInt32(index%,value%)
  int index=GI(2); if (index<0 || index>=nVars) return;
  *pVars[index]->pInt=GI(1);
}
DPF(TwWriteColor32){ // TwWriteColor32(index%,value%)
  int index=GI(2); if (index<0 || index>=nVars) return;
  *pVars[index]->pInt=GI(1);
}
DPF(TwWriteFloat){ // TwWriteFloat(index%,value#)
  int index=GI(2); if (index<0 || index>=nVars) return;
  *pVars[index]->pFloat=GF(1);
}
DPF(TwWriteDir3f){ // TwWriteDir3f(index%,value#())
  int index=GI(2); if (index<0 || index>=nVars) return;
  GFA(1,pVars[index]->pFloat,3);
}
DPF(TwWriteColor3f){ // TwWriteColor3f(index%,value#())
  int index=GI(2); if (index<0 || index>=nVars) return;
  GFA(1,pVars[index]->pFloat,3);
}
DPF(TwWriteColor4f){ // TwWriteColor4f(index%,value#())
  int index=GI(2); if (index<0 || index>=nVars) return;
  GFA(1,pVars[index]->pFloat,4);
}
DPF(TwWriteQuat4f){ // TwWriteQuat4f(index%,value#())
  int index=GI(2); if (index<0 || index>=nVars) return;
  GFA(1,pVars[index]->pFloat,4);
}

DPF(TwWriteCDString){ // TwWriteCDString(index%,value$)
  int index=GI(2); if (index<0 || index>=nVars) return;
  if (pVars[index]->pCDString){
    free(pVars[index]->pCDString);
    pVars[index]->pCDString=NULL;
  }
  pVars[index]->pCDString=strDup(GS(1));
}

// ############
// # UserData #
// ############
DPF(SetUserDataByIndex){ // SetUserDataByIndex(index%,UserData%)
  int index=GI(2); if (index<0 || index>=nVars) return;
  pVars[index]->UserData=GI(1);
}
DPF(GetUserDataByIndex){ // UserData% = GetUserDataByIndex(index%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  SIR(pVars[index]->UserData);
}

// ##############
// # Quaternion #
// ##############
DPF(MatrixFromQuat){ // Matrix#()() = MatrixFromQuat{QuatIndex%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  if (pVars[index]->VarType!=TW_TYPE_QUAT4F) return;
  float m[16];
  float a = pVars[index]->pFloat[0] * pVars[index]->pFloat[0]
          + pVars[index]->pFloat[1] * pVars[index]->pFloat[1]
          + pVars[index]->pFloat[2] * pVars[index]->pFloat[2]
          + pVars[index]->pFloat[3] * pVars[index]->pFloat[3]; 
  if (a>EPSILON) {
    a=sqrtf(a);
    float x = pVars[index]->pFloat[0]*a;
    float y = pVars[index]->pFloat[1]*a;
    float z = pVars[index]->pFloat[2]*a;
    float w = pVars[index]->pFloat[3]*a;
    float xx = x*x;
    float yy = y*y;
    float zz = z*z;
    float wx = w*x;
    float wy = w*y;
    float wz = w*z;
    float xy = x*y;
    float xz = x*z;
    float yz = y*z;
    m[0*4+0]=1.0f-2.0f*(yy+zz); m[0*4+1]=     2.0f*(xy+wz); m[0*4+2]=     2.0f*(xz-wy); m[0*4+3]=0.0f;
    m[1*4+0]=     2.0f*(xy-wz); m[1*4+1]=1.0f-2.0f*(xx+zz); m[1*4+2]=     2.0f*(yz+wx); m[1*4+3]=0.0f;
    m[2*4+0]=     2.0f*(xz+wy); m[2*4+1]=     2.0f*(yz-wx); m[2*4+2]=1.0f-2.0f*(xx+yy); m[2*4+3]=0.0f;
    m[3*4+0]=             0.0f; m[3*4+1]=             0.0f; m[3*4+2]=             0.0f; m[3*4+3]=1.0f;
  }
  else 
  {
    m[0*4+0]=1.0f; m[0*4+1]=0.0f; m[0*4+2]=0.0f; m[0*4+3]=0.0f;
    m[1*4+0]=0.0f; m[1*4+1]=1.0f; m[1*4+2]=0.0f; m[1*4+3]=0.0f;
    m[2*4+0]=0.0f; m[2*4+1]=0.0f; m[2*4+2]=1.0f; m[2*4+3]=0.0f;
    m[3*4+0]=0.0f; m[3*4+1]=0.0f; m[3*4+2]=0.0f; m[3*4+3]=1.0f;
  }
  SFAR2(m,4,4);
}

DPF(QuatFromMatrix) { // QuatFromMatrix(QuatIndex,Matix#()())
  int index=GI(2); if (index<0 || index>=nVars) return;
  if (pVars[index]->VarType!=TW_TYPE_QUAT4F) return;

  float m[16]; GFA2(1,m,4,4);
  float t = 0.25f*(1.0f+m[0*4+0]+m[1*4+1]+m[2*4+2]);

  if (t > EPSILON) {
    t = sqrtf(t);
    pVars[index]->pFloat[3] = t;
    t*= 4.0f;
    pVars[index]->pFloat[0] = (m[1*4+2]-m[2*4+1])/t;
    pVars[index]->pFloat[1] = (m[2*4+0]-m[0*4+2])/t;
    pVars[index]->pFloat[2] = (m[0*4+1]-m[1*4+0])/t;
  }
  else
  {
    pVars[index]->pFloat[3] = 0.0f;
    t = -0.5f*(m[1*4+1]+m[2*4+2]);
    if (t > EPSILON) {
      t = sqrtf(t);
      pVars[index]->pFloat[0] = t;
      t*= 2.0f;
      pVars[index]->pFloat[1] = m[0*4+1]/t;
      pVars[index]->pFloat[2] = m[0*4+2]/t;
    }
    else
    {
      pVars[index]->pFloat[0] = 0.0f;
      t = 0.5f*(1.0f-m[2*4+2]);
      if (t > EPSILON) {
        t = sqrtf(t);
        pVars[index]->pFloat[1] = t;
        pVars[index]->pFloat[2] = m[1*4+2]/(2.0f*t);
      }
      else
      {
        pVars[index]->pFloat[1] = 0.0f;
        pVars[index]->pFloat[2] = 1.0f;
      }
    }
  }
}

DPF(NormalizeQuat){ //NormalizeQuat(QuatIndex%)
  int index=GI(1); if (index<0 || index>=nVars) return;
  if (pVars[index]->VarType!=TW_TYPE_QUAT4F) return;
  float l = pVars[index]->pFloat[0] * pVars[index]->pFloat[0]
          + pVars[index]->pFloat[1] * pVars[index]->pFloat[1]
          + pVars[index]->pFloat[2] * pVars[index]->pFloat[2]
          + pVars[index]->pFloat[3] * pVars[index]->pFloat[3]; 
  if (l != 0.0f) {
    l=1.0f/sqrtf(l);
    pVars[index]->pFloat[0]*=l;
    pVars[index]->pFloat[1]*=l;
    pVars[index]->pFloat[2]*=l;
    pVars[index]->pFloat[3]*=l;
  }
  else
  {
    pVars[index]->pFloat[0]=0.0f;
    pVars[index]->pFloat[1]=0.0f;
    pVars[index]->pFloat[2]=0.0f;
    pVars[index]->pFloat[3]=1.0f;
  }
  
}
///////////////////////////////////////////////////////////////////////////////
//	AntTweakBarPlugin
bool DLLFUNC AntTweakBarPlugin::Load(IDLL_Basic4GL_FunctionRegistry &registry, bool isStandaloneExe)
{
  // initialise DLL stuff
  //isStandaloneExe=isStandaloneExe;
  // Fetch  interface
  gpCompiler=(IB4GLCompiler*)registry.FetchInterface("IB4GLCompiler",IB4GLCOMPILER_MAJOR,IB4GLCOMPILER_MINOR);

  RIC(TW_KMOD_NONE); RIC(TW_KMOD_SHIFT); RIC(TW_KMOD_CTRL); RIC(TW_KMOD_ALT); RIC(TW_KMOD_META);

  RIC(TW_KEY_BACKSPACE); RIC(TW_KEY_TAB); RIC(TW_KEY_CLEAR); RIC(TW_KEY_RETURN); RIC(TW_KEY_PAUSE);
  RIC(TW_KEY_ESCAPE); RIC(TW_KEY_SPACE); RIC(TW_KEY_DELETE); RIC(TW_KEY_UP); RIC(TW_KEY_DOWN); RIC(TW_KEY_RIGHT);
  RIC(TW_KEY_LEFT); RIC(TW_KEY_INSERT); RIC(TW_KEY_HOME); RIC(TW_KEY_END); RIC(TW_KEY_PAGE_UP); RIC(TW_KEY_PAGE_DOWN);
  RIC(TW_KEY_F1); RIC(TW_KEY_F2); RIC(TW_KEY_F3); RIC(TW_KEY_F4); RIC(TW_KEY_F5); RIC(TW_KEY_F6);
  RIC(TW_KEY_F7); RIC(TW_KEY_F8); RIC(TW_KEY_F9); RIC(TW_KEY_F10); RIC(TW_KEY_F11); RIC(TW_KEY_F12);
  
  RIC(TW_MOUSE_RELEASED); RIC(TW_MOUSE_PRESSED);
  RIC(TW_MOUSE_LEFT); RIC(TW_MOUSE_MIDDLE); RIC(TW_MOUSE_RIGHT);

  RIC(TW_TYPE_UNDEF);
  RIC(TW_TYPE_BOOL32);     // boolean
  RIC(TW_TYPE_INT32);      // integer
  RIC(TW_TYPE_COLOR32);    // 32 bits color.

  RIC(TW_TYPE_FLOAT);      // float
  
  RIC(TW_TYPE_DIR3F);      // direction vector represented by 3 floats
  RIC(TW_TYPE_COLOR3F);    // 3 floats color. Order is RGB.

  RIC(TW_TYPE_COLOR4F);    // 4 floats color. Order is RGBA.
  RIC(TW_TYPE_QUAT4F);     // 4 floats encoding a quaternion {qx,qy,qz,qs}

  RIC(TW_TYPE_CDSTRING);   // Null-terminated C Dynamic String (pointer to an array of char dynamically allocated with malloc/realloc/strdup)
  

  hBar = RT(TweakBar,0,1);
    RTIF(Address);
  // #######
  // # API #
  // #######  
  RSF(TwGetLastError);
  RIF(TwInit);AIP;AIP;
  RIF(TwIsInit);
  RIF(TwTerminate);
  
  RIF(TwDefine);ASP;

  RIF(TwDraw);

  RIF(TwWindowSize);AIP;AIP;
  RIF(TwMouseMotion);AFP;AFP;
  RIF(TwMouseButton);AIP;AIP;
  RIF(TwMouseButtons);AIP;AIP;AIP;
  RIF(TwMouseWheel);AIP;
  RIF(TwKeyPressed);AIP;AIP;
  RIF(TwScanKey);AIP;

  // ############
  // # Tweakbar #
  // ############
  RTF(TwNewBar,hBar);ASP;
  RIF(TwDeleteBar);ATPR(hBar);
  RIF(TwDeleteAllBars);

  RIF(TwSetTopBar);ATPR(hBar);
  RTF(TwGetTopBar,hBar);
  
  RIF(TwSetBottomBar);ATPR(hBar);
  RTF(TwGetBottomBar,hBar);
  
  RSF(TwGetBarName);ATPR(hBar);
  RTF(TwGetBarByName,hBar);ASP;

  RIF(TwGetBarCount);
  RTF(TwGetBarByIndex,hBar);AIP;
  
  RIF(TwRefreshBar);ATPR(hBar);

  // ##########
  // # Button #
  // ##########
  RIF2(TwAddButton,TwAddButtonD); ATP(hBar);ASP; ASP; ASP;
  RIF(TwAddButton);               ATP(hBar);ASP; ASP;
  
  // #############
  // # Separator #
  // #############
  RIF2(TwAddSeparator,TwAddSeparatorD); ATP(hBar);ASP; ASP;
  RIF(TwAddSeparator);                  ATP(hBar);ASP;

  // ################
  // # CallbackVars #
  // ################
  RIF2(TwAddVarCB,TwAddVarCBD);       ATP(hBar);ASP; AIP; ASP;ASP; ASP;
  RIF2(TwAddVarCB,TwAddVarCB);        ATP(hBar);ASP; AIP; ASP;ASP;

  RIF2(TwAddVarROCB,TwAddVarROCBD);   ATP(hBar);ASP; AIP; ASP;     ASP;
  RIF2(TwAddVarROCB,TwAddVarROCB);    ATP(hBar);ASP; AIP; ASP;

  // ###########
  // # RO Vars #
  // ###########
  RIF2(TwAddIntegerVarRO,TwAddVarROD); ATP(hBar);ASP;AIP; AIP;     ASP;
  RIF2(TwAddSingleVarRO ,TwAddVarROD); ATP(hBar);ASP;AIP; AFP;     ASP;
  RIF2(TwAddArrayVarRO  ,TwAddVarROD); ATP(hBar);ASP;AIP; AFAP(1); ASP;
  RIF2(TwAddStringVarRO ,TwAddVarROD); ATP(hBar);ASP;AIP; ASP;     ASP;

  RIF2(TwAddIntegerVarRO,TwAddVarRO); ATP(hBar);ASP;AIP; AIP;
  RIF2(TwAddSingleVarRO ,TwAddVarRO); ATP(hBar);ASP;AIP; AFP;
  RIF2(TwAddArrayVarRO  ,TwAddVarRO); ATP(hBar);ASP;AIP; AFAP(1);
  RIF2(TwAddStringVarRO ,TwAddVarRO); ATP(hBar);ASP;AIP; ASP;

  RIF2(TwAddIntegerVarRW,TwAddVarRWD); ATP(hBar);ASP;AIP; AIP;     ASP;
  RIF2(TwAddSingleVarRW ,TwAddVarRWD); ATP(hBar);ASP;AIP; AFP;     ASP;
  RIF2(TwAddArrayVarRW  ,TwAddVarRWD); ATP(hBar);ASP;AIP; AFAP(1); ASP;
  RIF2(TwAddStringVarRW ,TwAddVarRWD); ATP(hBar);ASP;AIP; ASP;     ASP;

  RIF2(TwAddIntegerVarRW,TwAddVarRW); ATP(hBar);ASP;AIP; AIP;
  RIF2(TwAddSingleVarRW ,TwAddVarRW); ATP(hBar);ASP;AIP; AFP;
  RIF2(TwAddArrayVarRW  ,TwAddVarRW); ATP(hBar);ASP;AIP; AFAP(1);
  RIF2(TwAddStringVarRW ,TwAddVarRW); ATP(hBar);ASP;AIP; ASP;

  // ########
  // # read #
  // ########
  RIF(TwReadBool32);    AIP;
  RIF(TwReadInt32);     AIP;
  RIF(TwReadColor32);   AIP;
  RFF(TwReadFloat);     AIP;
  RSF(TwReadCDString);  AIP;
  RFAF(TwReadDir3f,1);  AIP;
  RFAF(TwReadColor3f,1);AIP;
  RFAF(TwReadColor4f,1);AIP;
  RFAF(TwReadQuat4f,1); AIP;

  // #########
  // # write #
  // #########
  RVF(TwWriteBool32);  AIP;AIP;
  RVF(TwWriteInt32);   AIP;AIP;
  RVF(TwWriteColor32); AIP;AIP;
  RVF(TwWriteFloat);   AIP;AFP;
  RVF(TwWriteCDString);AIP;ASP;
  RVF(TwWriteDir3f);   AIP;AFAP(1);
  RVF(TwWriteColor3f); AIP;AFAP(1);
  RVF(TwWriteColor4f); AIP;AFAP(1);
  RVF(TwWriteQuat4f);  AIP;AFAP(1);
  
  // ############
  // # userdata #
  // ############
  RVF(SetUserDataByIndex);AIP;AIP;
  RIF(GetUserDataByIndex);AIP;

  // ##############
  // # Quaternion #
  // ##############
  RFAF(MatrixFromQuat,2);AIP; 
  RVF(QuatFromMatrix);AIP;AFAP(2);
  RVF(NormalizeQuat);AIP;
  
  return true;
}

void DLLFUNC AntTweakBarPlugin::Unload() {
  TwTerminate(); IsInit=0;
  FreeAllVars();
  delete this;
}

bool DLLFUNC AntTweakBarPlugin::Start(){
  //hWin = FindWindow("gbOPENGL",NULL);
  FreeAllVars();
  return true;
}

void DLLFUNC AntTweakBarPlugin::End(){
  TwTerminate(); IsInit=0;
  FreeAllVars();
}

void DLLFUNC AntTweakBarPlugin::Pause(){
    // Called when the program is paused (i.e. during debugging).
    // A full screen window might hide itself here so that the user can see the
    // Basic4GL edit window (and debug their program)
}

void DLLFUNC AntTweakBarPlugin::Resume(){
    // Called when the program is resumed (i.e. during debugging).
    // This is also called when the user "steps"
}

void DLLFUNC AntTweakBarPlugin::DelayedResume(){
    // Called when the program is resumed for more than a short "step".
    // This is where a fullscreen window would re-show itself (if it did so in ::Resume()
    // we would get flickering when the user steps through the program)
}

void DLLFUNC AntTweakBarPlugin::GetError(char *error){
  if (pluginError != NULL) {
    for (int i = 0; pluginError[i] != 0; i++){
      error[i] = pluginError[i];
   }
 }
}

void DLLFUNC AntTweakBarPlugin::ProcessMessages(){
  // Called periodically (in practice, quite frequently).
  // This would be a good place to process windows messages etc.
}

///////////////////////////////////////////////////////////////////////////////
// DLL exported functions
extern "C" {

  // Query function
  DLLExport int DLLFUNC Basic4GL_Query(char *details, int *major, int *minor) {
    char detailSrc[] = "for Basic4GL";
    for (int i = 0; detailSrc[i] != 0; i++) {
      details[i] = detailSrc[i];
    }
    *major = 1;
    *minor = 16;
    return BASIC4GL_DLL_VERSION;
  }

  DLLExport IDLL_Basic4GL_Plugin * DLLFUNC Basic4GL_Init(){
    return new AntTweakBarPlugin();
  }
}